@@ -1,17 +1,17 @@
 ;;; borg.el --- Assimilate Emacs packages as Git submodules  -*- lexical-binding:t -*-
 
-;; Copyright (C) 2016-2025 Jonas Bernoulli
+;; Copyright (C) 2016-2026 Jonas Bernoulli
 
 ;; Author: Jonas Bernoulli <emacs.borg@jonas.bernoulli.dev>
 ;; Homepage: https://github.com/emacscollective/borg
 ;; Keywords: tools
 
-;; Package-Version: 4.3.0
-;; Package-Revision: v4.3.0-0-g28e6ba50fab8
+;; Package-Version: 4.4.0
+;; Package-Revision: v4.4.0-0-g61f9e083b8d9
 ;; Package-Requires: (
-;;     (emacs "27.1")
+;;     (emacs "28.1")
 ;;     (epkg   "4.1")
-;;     (magit  "4.4"))
+;;     (magit  "4.5"))
 
 ;; SPDX-License-Identifier: GPL-3.0-or-later
 
@@ -46,6 +46,62 @@
 
 ;;; Code:
 
+(put 'if-let 'byte-obsolete-info nil)
+(put 'when-let 'byte-obsolete-info nil)
+
+(defun borg-report-load-duration (&optional file format-string)
+  "Report how long it takes to load the file currently being loaded.
+
+This function should be called very early in the file, ideally
+before doing anything else.  Print two messages; one right now
+and another once the file has been loaded.
+
+If optional FILE is non-nil, instead report how long it takes to
+load that file (this assumes that it is currently being loaded).
+Optional FORMAT-STRING, if non-nil, is passed to `message'.  It
+should end with \"...\" and contain zero or one %s, which stands
+for the file name."
+  (when load-file-name
+    (letrec ((start (current-time))
+             (load-file (or file load-file-name))
+             (format (or format-string "Loading %s..."))
+             (fn (lambda (file)
+                   (when (equal file load-file)
+                     (message "%s%s"
+                              (format-message format load-file)
+                              (format-message "done (%.3fs)"
+                                              (float-time
+                                               (time-subtract
+                                                (current-time) start))))
+                     (remove-hook 'after-load-functions fn)))))
+      (add-hook 'after-load-functions fn 98)
+      (message format load-file))))
+
+(when user-init-file
+  (unless after-init-time
+    (borg-report-load-duration user-init-file nil))
+  (borg-report-load-duration))
+
+(defvar borg-debug-user-init-files t
+  "Whether to extend the effect of `--debug-init' to user init files.
+If non-nil and `--debug-init' is used, enable `debug-on-error' and
+set `use-package-verbose' to `debug'.  Reset these variables using
+`after-init-hook'.")
+
+(when (and borg-debug-user-init-files
+           init-file-debug
+           (not after-init-time))
+  (setq debug-on-error 'enabled-by-borg)
+  (defvar use-package-verbose)
+  (setq use-package-verbose 'debug)
+  (letrec ((fn (lambda ()
+                 (when (eq debug-on-error 'enabled-by-borg)
+                   (setq debug-on-error nil))
+                 (when (eq use-package-verbose 'debug)
+                   (setq use-package-verbose nil))
+                 (remove-hook 'after-init-hook fn))))
+    (add-hook 'after-init-hook fn 98)))
+
 (require 'bytecomp)
 (require 'cl-lib)
 (require 'comp nil t)
@@ -86,8 +142,6 @@ enclosed in a `progn' form.  ELSE-FORMS may be empty."
 
 (defvar inhibit-message-regexps) ; added in Emacs 29.1
 
-(defvar byte+native-compile) ; for Emacs 27
-(defvar comp-files-queue)    ; for Emacs 27
 (defvar native-comp-eln-load-path)
 (defvar native-compile-target-directory)
 
@@ -444,6 +498,13 @@ to variable `borg-rewrite-urls-alist' (which see)."
   (require 'epkg nil t)
   (completing-read prompt (borg-clones) nil t nil 'epkg-package-history))
 
+(defun borg--load-config (file &optional format-string nomessage)
+  (let ((file (expand-file-name file borg-user-emacs-directory)))
+    (when (file-exists-p file)
+      (unless nomessage
+        (borg-report-load-duration file format-string))
+      (load file nil t t))))
+
 ;;; Activation
 
 (defun borg-initialize ()
@@ -461,24 +522,20 @@ If Emacs is running without an interactive terminal, then first
 load \"`borg-user-emacs-directory'/etc/borg/init.el\", if that
 exists."
   (when noninteractive
-    (let ((init (convert-standard-filename
-                 (expand-file-name "etc/borg/init.el"
-                                   borg-user-emacs-directory))))
-      (when (file-exists-p init)
-        (load-file init))))
+    (borg--load-config "etc/borg/init.el"))
   (info-initialize)
   (let ((start (current-time))
         (skipped 0)
         (initialized 0))
     (borg-do-drones (drone)
       (cond
-       ((borg--drone-disabled-p drone)
-        (cl-incf skipped))
-       ((not (file-exists-p (borg-worktree drone)))
-        (cl-incf skipped))
-       (t
-        (cl-incf initialized)
-        (borg-activate drone))))
+        ((borg--drone-disabled-p drone)
+         (cl-incf skipped))
+        ((not (file-exists-p (borg-worktree drone)))
+         (cl-incf skipped))
+        (t
+         (cl-incf initialized)
+         (borg-activate drone))))
     (let* ((message (current-message))
            (inhibit (and message
                          (string-match-p
@@ -506,7 +563,7 @@ if it exists."
          (let ((file (expand-file-name (format "%s-%s.el" clone part) dir)))
            (and (file-exists-p file)
                 (with-demoted-errors "Error loading autoloads: %s"
-                  (load file nil t))
+                  (load file nil t t))
                 ;; We cannot rely on the autoloads file doing that.
                 (add-to-list 'load-path dir)))))
     (dolist (dir (borg-load-path clone))
@@ -538,7 +595,7 @@ otherwise."
     (error "borg-batch-rebuild is to be used only with --batch"))
   (borg-do-drones (drone)
     (unless (and quick (borg-get-all drone "build-step"))
-      (borg--remove-autoloads drone quick)))
+      (borg--remove-autoloads drone)))
   (when (borg-dronep "org")
     ;; `org-loaddefs.el' has to exist when compiling a library
     ;; which depends on `org', else we get warnings about that
@@ -549,22 +606,18 @@ otherwise."
     (message "\n--- [%s] ---\n" drone)
     (let ((dir (borg-worktree drone)))
       (cond
-       ((borg--drone-disabled-p drone)
-        (message "Skipped (Disabled)"))
-       ((let ((min (cdr (assoc drone borg-minimal-emacs-alist))))
-          (and min (version< emacs-version min)
-               (message "Skipped (Requires Emacs >= %s)" min))))
-       ((not (file-exists-p dir))
-        (message "Skipped (Missing)"))
-       ((static-if (functionp 'directory-empty-p) ;since Emacs 28.1
-            (directory-empty-p dir)
-          (and (file-directory-p dir)
-               (null (directory-files
-                      dir nil directory-files-no-dot-files-regexp t))))
-        (message "Skipped (Empty directory)"))
-       ((and quick (borg-get-all drone "build-step"))
-        (message "Skipped (Expensive to build)"))
-       ((borg-build drone nil native)))))
+        ((borg--drone-disabled-p drone)
+         (message "Skipped (Disabled)"))
+        ((let ((min (cdr (assoc drone borg-minimal-emacs-alist))))
+           (and min (version< emacs-version min)
+                (message "Skipped (Requires Emacs >= %s)" min))))
+        ((not (file-exists-p dir))
+         (message "Skipped (Missing)"))
+        ((directory-empty-p dir)
+         (message "Skipped (Empty directory)"))
+        ((and quick (borg-get-all drone "build-step"))
+         (message "Skipped (Expensive to build)"))
+        ((borg-build drone nil native)))))
   (borg-batch-rebuild-init))
 
 (defun borg-batch-rebuild-init ()
@@ -579,7 +632,6 @@ This function is to be used only with `--batch'."
                             (concat (user-real-login-name) ".el"))))
       (when (file-exists-p file)
         (message "\n--- [%s] ---\n" file)
-        (load-file file)
         (byte-recompile-file (expand-file-name file) t 0)))))
 
 (defun borg-build (clone &optional activate native)
@@ -593,37 +645,35 @@ and optional NATIVE are both non-nil, then also compile natively."
     (setq clone (substring clone 0 -1)))
   (borg-clean clone)
   (cond
-   (noninteractive
-    (let ((borg-compile-function borg-compile-function))
-      (when (and (fboundp 'native-comp-available-p)
-                 (native-comp-available-p))
-        (when native
-          (setq borg-compile-function
-                (if (functionp native) native #'borg-byte+native-compile)))
-        (when (memq borg-compile-function '( borg--native-compile
-                                             native-compile
-                                             native-compile-async))
-          (message "WARNING: Using `%s' instead of unsuitable `%s'"
-                   'borg-byte+native-compile borg-compile-function)
-          (setq borg-compile-function #'borg-byte+native-compile)))
-      (borg--build-noninteractive clone)
-      (when activate
-        (borg-activate clone))))
-   ((let ((process (borg--build-interactive clone)))
-      (when activate
-        (add-function :after (process-sentinel process)
-                      (lambda (_process event)
-                        (when (string-equal event "finished\n")
-                          (borg-activate clone)))))))))
+    (noninteractive
+     (let ((borg-compile-function borg-compile-function))
+       (when (and (fboundp 'native-comp-available-p)
+                  (native-comp-available-p))
+         (when native
+           (setq borg-compile-function
+                 (if (functionp native) native #'borg-byte+native-compile)))
+         (when (memq borg-compile-function '( borg--native-compile
+                                              native-compile
+                                              native-compile-async))
+           (message "WARNING: Using `%s' instead of unsuitable `%s'"
+                    'borg-byte+native-compile borg-compile-function)
+           (setq borg-compile-function #'borg-byte+native-compile)))
+       (borg--build-noninteractive clone)
+       (when activate
+         (borg-activate clone))))
+    ((let ((process (borg--build-interactive clone)))
+       (when activate
+         (add-function :after (process-sentinel process)
+                       (lambda (_process event)
+                         (when (string-equal event "finished\n")
+                           (borg-activate clone)))))))))
 
 (defun borg--build-noninteractive (clone)
   (let ((default-directory (borg-worktree clone))
         (build-cmd (if (functionp borg-build-shell-command)
                        (funcall borg-build-shell-command clone)
-                     borg-build-shell-command))
-        (config (borg--config-file)))
-    (when (file-exists-p config)
-      (load config nil t t))
+                     borg-build-shell-command)))
+    (borg--load-config "etc/borg/config.el")
     (if-let ((commands (borg-get-all clone "build-step")))
         (borg--run-build-commands clone commands build-cmd)
       (let ((path (mapcar #'file-name-as-directory (borg-load-path clone))))
@@ -668,19 +718,14 @@ and optional NATIVE are both non-nil, then also compile natively."
              (and file
                   (string-match-p emacs-lisp-file-regexp file)
                   (file-in-directory-p file top))))))
+  (borg--load-config "etc/borg/config.el")
   (let ((buffer (get-buffer-create "*Borg Build*"))
-        (config (borg--config-file))
         (process-connection-type nil))
     (with-current-buffer buffer
       (setq default-directory borg-user-emacs-directory)
       (borg-build-mode)
       (goto-char (point-max))
       (let ((inhibit-read-only t))
-        (when (file-exists-p config)
-          (insert (format "\n(%s) Loading %s\n\n"
-                          (format-time-string "%H:%M:%S")
-                          config))
-          (load-file config))
         (unless (looking-back "\n\n" (- (point) 2))
           (insert ?\n))
         (insert (format "(%s) Building %s\n\n"
@@ -766,9 +811,9 @@ and optional NATIVE are both non-nil, then also compile natively."
                        (funcall fn string nil type))
                      '((name . "no-message")))
          ,bodyform)
-      (advice-remove 'progress-reporter-do-update #'ignore)
-      (advice-remove 'progress-reporter-done #'ignore)
-      (advice-remove 'byte-compile-info "no-message")))
+     (advice-remove 'progress-reporter-do-update #'ignore)
+     (advice-remove 'progress-reporter-done #'ignore)
+     (advice-remove 'byte-compile-info "no-message")))
 
 (defun borg-update-autoloads (clone &optional path)
   "Update autoload files for the clone named CLONE in the directories in PATH."
@@ -802,7 +847,8 @@ and optional NATIVE are both non-nil, then also compile natively."
           (when-let ((buf (find-buffer-visiting file)))
             (kill-buffer buf)))
 
-      ;; For Emacs < 29.1.
+      ;; For Emacs 28.
+      (borg--remove-autoloads clone)
       (let ((backup-inhibited t)
             (version-control 'never)
             (noninteractive t)
@@ -813,12 +859,10 @@ and optional NATIVE are both non-nil, then also compile natively."
         ;; caused #128 and to prevent that we use `with-temp-buffer'.
         (with-temp-buffer
           (let ((coding-system-for-write 'utf-8-emacs-unix))
-            (write-region (and (functionp 'autoload-rubric)
-                               (autoload-rubric file "package" t))
+            (write-region (autoload-rubric file "package" t)
                           nil file nil 'silent)))
         (borg--silence-loaddefs-generate
-          (let ((generated-autoload-file file))
-            (apply 'update-directory-autoloads path))))
+          (make-directory-autoloads path file)))
       (when-let ((buf (find-buffer-visiting file)))
         (kill-buffer buf)))))
 
@@ -887,36 +931,36 @@ and optional NATIVE are both non-nil, then also compile natively."
 
 (defun borg-byte+native-compile (file)
   (cond
-   ((or (equal (getenv "NATIVE_DISABLED") "1")
-        (member (file-name-nondirectory file)
-                borg-native-compile-deny-list))
-    (byte-compile-file file))
-   ((and (fboundp 'comp--native-compile)
-         (fboundp 'comp-ensure-native-compiler))
-    (comp-ensure-native-compiler)
-    (let* ((byte+native-compile t)
-           (byte-to-native-output-buffer-file nil)
-           (native-compile-target-directory (car native-comp-eln-load-path))
-           (eln-file (comp--native-compile file)))
-      (pcase byte-to-native-output-buffer-file
-        (`(,temp-buffer . ,target-file)
-         (unwind-protect
-             (progn
-               (borg--byte-write-target-file temp-buffer target-file)
-               (when (stringp eln-file)
-                 (set-file-times eln-file)))
-           (kill-buffer temp-buffer)
-           (file-exists-p target-file))))))
-   ((error "Emacs %s does not support native compilation" emacs-version))))
+    ((or (equal (getenv "NATIVE_DISABLED") "1")
+         (member (file-name-nondirectory file)
+                 borg-native-compile-deny-list))
+     (byte-compile-file file))
+    ((and (fboundp 'comp--native-compile)
+          (fboundp 'comp-ensure-native-compiler))
+     (comp-ensure-native-compiler)
+     (let* ((byte+native-compile t)
+            (byte-to-native-output-buffer-file nil)
+            (native-compile-target-directory (car native-comp-eln-load-path))
+            (eln-file (comp--native-compile file)))
+       (pcase byte-to-native-output-buffer-file
+         (`(,temp-buffer . ,target-file)
+          (unwind-protect
+              (progn
+                (borg--byte-write-target-file temp-buffer target-file)
+                (when (stringp eln-file)
+                  (set-file-times eln-file)))
+            (kill-buffer temp-buffer)
+            (file-exists-p target-file))))))
+    ((error "Emacs %s does not support native compilation" emacs-version))))
 
 (defun borg-byte+native-compile-async (file)
   (byte-compile-file file)
   (cond
-   ((and (fboundp 'native-compile-async)
-         (fboundp 'comp-ensure-native-compiler))
-    (comp-ensure-native-compiler)
-    (native-compile-async file))
-   ((error "Emacs %s does not support native compilation" emacs-version))))
+    ((and (fboundp 'native-compile-async)
+          (fboundp 'comp-ensure-native-compiler))
+     (comp-ensure-native-compiler)
+     (native-compile-async file))
+    ((error "Emacs %s does not support native compilation" emacs-version))))
 
 (if (fboundp 'byte-write-target-file)
     (defalias 'borg--byte-write-target-file 'byte-write-target-file)
@@ -935,11 +979,7 @@ and optional NATIVE are both non-nil, then also compile natively."
               (cons (lambda () (ignore-errors (delete-file tempfile)))
                     kill-emacs-hook)))
         (unless (= temp-modes desired-modes)
-          (with-no-warnings
-            ;; Native compilation, and thus this function, may only
-            ;; be used for Emacs > 28.1.  So don't warn about third
-            ;; argument that does exist in older releases.
-            (set-file-modes tempfile desired-modes 'nofollow)))
+          (set-file-modes tempfile desired-modes 'nofollow))
         (write-region (point-min) (point-max) tempfile nil 1)
         (if byte-native-compiling
             (setf byte-to-native-output-buffer-file
@@ -1016,10 +1056,9 @@ non-nil, then try those files instead."
           (shell-command cmd))
         (message "  Running `%s'...done" cmd)))))
 
-(defun borg--remove-autoloads (drone &optional quick)
+(defun borg--remove-autoloads (drone)
   (unless (or (borg--drone-disabled-p drone)
-              (not (file-exists-p (borg-worktree drone)))
-              (and quick (borg-get-all drone "build-step")))
+              (not (file-exists-p (borg-worktree drone))))
     (dolist (dir (borg-load-path drone))
       (dolist (file (directory-files
                      dir t "\\(\\.elc\\|-autoloads\\.el\\|-loaddefs\\.el\\)\\'"
@@ -1060,8 +1099,8 @@ with a prefix argument, then also read the url in the minibuffer.
 With a negative prefix argument only add the submodule but don't
 build and activate the drone."
   (interactive
-   (nconc (borg-read-package "Assimilate package: " current-prefix-arg)
-          (list (< (prefix-numeric-value current-prefix-arg) 0))))
+    (nconc (borg-read-package "Assimilate package: " current-prefix-arg)
+           (list (< (prefix-numeric-value current-prefix-arg) 0))))
   (borg--maybe-confirm-unsafe-action "assimilate" package url)
   (message "Assimilating %s..." package)
   (let ((default-directory borg-top-level-directory))
@@ -1183,6 +1222,31 @@ Formatting is according to the commit message conventions."
      (process-lines "git" "diff-index" "--name-status" "--cached" "HEAD"
                     "--" (file-relative-name borg-drones-directory)))))
 
+(defun borg-report-init-duration ()
+  "Report how long Emacs initialization and `after-init-hook' took.
+This adds functions to `after-init-hook' to determine how long it
+takes to run that, and to report those findings and also how long
+Emacs initialization took.  If that hook has already run, do nothing."
+  (unless after-init-time
+    (letrec ((before nil)
+             (beg  (lambda ()
+                     (setq before (current-time))
+                     (message "Running after-init-hook...")
+                     (remove-hook 'after-init-hook beg)))
+             (end  (lambda ()
+                     (message "Running after-init-hook...done (%.3fs)"
+                              (float-time (time-subtract (current-time)
+                                                         before)))
+                     (remove-hook 'after-init-hook end)))
+             (init (lambda ()
+                     (message "Total init time %.3fs"
+                              (float-time (time-subtract after-init-time
+                                                         before-init-time)))
+                     (remove-hook 'after-init-hook init))))
+      (add-hook 'after-init-hook beg  -98)
+      (add-hook 'after-init-hook end  +98)
+      (add-hook 'after-init-hook init +98))))
+
 ;;; Integrations
 
 (defun borg-propertize-module-path (path)
@@ -1202,10 +1266,6 @@ Formatting is according to the commit message conventions."
           (match-string 0 ver)
         (error "Cannot determine Git version (from %s)" ver)))))
 
-(defun borg--config-file ()
-  (convert-standard-filename
-   (expand-file-name "etc/borg/config.el" borg-user-emacs-directory)))
-
 (defun borg--maybe-absorb-gitdir (pkg)
   (if (version< (borg--git-version) "2.12.0")
       (let ((default-directory (borg-worktree pkg))
@@ -1337,7 +1397,4 @@ Do you really want to %s it? "))))
 
 ;;; _
 (provide 'borg)
-;; Local Variables:
-;; indent-tabs-mode: nil
-;; End:
 ;;; borg.el ends here
